/*
 * 
 * Name:   ZkHelper
 * 
 * Author: Bhaskar S
 * 
 * Date:   05/24/2014
 * 
 */

package com.polarsparc.zookeeper;

import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.CountDownLatch;
import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.zookeeper.AsyncCallback.StringCallback;
import org.apache.zookeeper.AsyncCallback.StatCallback;
import org.apache.zookeeper.AsyncCallback.ChildrenCallback;
import org.apache.zookeeper.AsyncCallback.DataCallback;
import org.apache.zookeeper.AsyncCallback.VoidCallback;
import org.apache.zookeeper.CreateMode;
import org.apache.zookeeper.data.Stat;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.Watcher;
import org.apache.zookeeper.WatchedEvent;
import org.apache.zookeeper.ZooDefs;
import org.apache.zookeeper.ZooKeeper;

public final class ZkHelper {
    private static final Logger log = LoggerFactory.getLogger(ZkHelper.class);
    
    // ----- Setup Connection to ZooKeeper -----
    
    final static ZooKeeper connect(String hostPortList, int sessionTimeout)
        throws Exception {
        CountDownLatch signal = new CountDownLatch(1);
        
        ZooKeeper zk = new ZooKeeper(hostPortList, sessionTimeout, (WatchedEvent event) -> {
            if (event.getState() == Watcher.Event.KeeperState.SyncConnected) {
                log.info("ZK[" + hostPortList + "] -> Connected !!!");
                signal.countDown();
            }
        });
        
        signal.await();
        
        return zk;
    }
    
    // ----- Create a Znode (Persistent or Ephemeral) -----
    
    final static boolean createZnode(ZooKeeper zk, CreateMode mode, String path, byte[] data) {
        boolean success = false;
        
        try {
            zk.create(path,
                      data,
                      ZooDefs.Ids.OPEN_ACL_UNSAFE,
                      mode);
            success = true;
        }
        catch (InterruptedException | KeeperException ex) {
            log.error("createZnode: Exception", ex);
        }
        
        return success;
    }
    
    final static boolean asyncCreateZnode(ZooKeeper zk, CreateMode mode, String path, byte[] data) {
        CountDownLatch signal = new CountDownLatch(1);
        
        class CreateZnodeContext {
            boolean success = false;
            CreateMode type;
            byte[] data;

            CreateZnodeContext(CreateMode type, byte[] data) {
                this.type = type;
                this.data = data;
            }
            
            void setSuccess() {
                success = true;
            }
        }
        
        final CreateZnodeContext ctx = new CreateZnodeContext(mode, data);
        
        class CreateZnodeStringCallback implements StringCallback {
            @Override
            public void processResult(int rc, String path, Object ctx, String name) {
                switch(KeeperException.Code.get(rc)) {
                    case OK:
                    {
                        ((CreateZnodeContext)ctx).setSuccess();
                        log.info("ZK[" + path + "] -> Created !!!");
                        signal.countDown();
                        break;
                    }
                    case NODEEXISTS:
                    {
                        log.info("ZK[" + path + "] -> Already Exists !!!");
                        signal.countDown();
                        break;
                    }
                    case CONNECTIONLOSS:
                    case SESSIONEXPIRED:
                    case SESSIONMOVED:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc) + ", Will Retry !!!");
                        try {
                            zk.create(path,
                                      ((CreateZnodeContext)ctx).data,
                                      ZooDefs.Ids.OPEN_ACL_UNSAFE,
                                      ((CreateZnodeContext)ctx).type,
                                      new CreateZnodeStringCallback(),
                                      ctx);
                        }
                        catch (Exception ex) {
                            log.error("asyncCreateZnode: Exception", ex);
                        }
                        break;
                    }
                    default:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc) + ", Failed !!!");
                        signal.countDown();
                        break;
                    }
                }
            }
        }
        
        zk.create(path,
                  ctx.data,
                  ZooDefs.Ids.OPEN_ACL_UNSAFE,
                  ctx.type,
                  new CreateZnodeStringCallback(),
                  ctx);
        
        try {
            signal.await();
        }
        catch (InterruptedException ex) {
        }
        
        return ctx.success;
    }
    
    // ----- Check for Znode Existence and Get Stats -----
    
    final static boolean checkZnode(ZooKeeper zk, String path, Watcher watcher) {
        boolean success = false;
        
        try {
            Stat stat = zk.exists(path,
                                  watcher);
            if (stat != null) {
                success = true;
                log.info("ZK[" + path + "] -> Exists, Stat: {ctime = " + new Date(stat.getCtime())
                    + ", mtime = " + new Date(stat.getMtime())
                    + ", version = " + stat.getVersion() + "}");
            }
        }
        catch (InterruptedException | KeeperException ex) {
            log.error("checkZnode: Exception", ex);
        }
        
        return success;
    }
    
    final static boolean asyncCheckZnode(ZooKeeper zk, String path, Watcher watcher) {
        CountDownLatch signal = new CountDownLatch(1);
        
        class CheckZnodeContext {
            boolean success = false;
            
            void setSuccess() {
                success = true;
            }
        }
        
        final CheckZnodeContext ctx = new CheckZnodeContext();
        
        class CheckZnodeStatCallback implements StatCallback {
            @Override
            public void processResult(int rc, String path, Object ctx, Stat stat) {
                switch(KeeperException.Code.get(rc)) {
                    case OK:
                    {
                        ((CheckZnodeContext)ctx).setSuccess();
                        log.info("ZK[" + path + "] -> Exists, Stat: {ctime = " + new Date(stat.getCtime())
                            + ", mtime = " + new Date(stat.getMtime())
                            + ", version = " + stat.getVersion() + "}");
                        signal.countDown();
                        break;
                    }
                    case NONODE:
                    {
                        log.info("ZK[" + path + "] -> Does not Exists !!!");
                        signal.countDown();
                        break;
                    }
                    case CONNECTIONLOSS:
                    case SESSIONEXPIRED:
                    case SESSIONMOVED:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc));
                        try {
                            zk.exists(path,
                                      watcher,
                                      new CheckZnodeStatCallback(),
                                      ctx);
                        }
                        catch (Exception ex) {
                            log.error("asyncCheckZnode: Exception", ex);
                        }
                        break;
                    }
                    default:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc) + ", Failed !!!");
                        signal.countDown();
                        break;
                    }
                }
            }
        }
        
        zk.exists(path,
                  watcher,
                  new CheckZnodeStatCallback(),
                  ctx);
        
        try {
            signal.await();
        }
        catch (InterruptedException ex) {
        }
        
        return ctx.success;
    }
    
    // ----- Get Children from Existing Znode -----
    
    final static List<String> getZnodeChildren(ZooKeeper zk, String path, Watcher watcher) {
        List<String> children = new ArrayList<>();
        
        try {
            List<String> list = zk.getChildren(path,
                                               watcher);
            if (list != null) {
                children.addAll(list);
            }
        }
        catch (InterruptedException | KeeperException ex) {
            log.error("getZnodeChildren: Exception", ex);
        }
        
        return children;
    }
    
    final static List<String> asyncGetZnodeChildren(ZooKeeper zk, String path, Watcher watcher) {
        CountDownLatch signal = new CountDownLatch(1);
        
        final List<String> children = new ArrayList<>();
        
        class GetZnodeChildrenCallback implements ChildrenCallback {
            @Override
            public void processResult(int rc, String path, Object ctx, List<String> list) {
                switch(KeeperException.Code.get(rc)) {
                    case OK:
                    {
                        if (list != null) {
                            children.addAll(list);
                        }
                        log.info("ZK[" + path + "] -> Children: " + list);
                        signal.countDown();
                        break;
                    }
                    case NONODE:
                    {
                        log.info("ZK[" + path + "] -> Does not Exists !!!");
                        signal.countDown();
                        break;
                    }
                    case CONNECTIONLOSS:
                    case SESSIONEXPIRED:
                    case SESSIONMOVED:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc));
                        try {
                            zk.getChildren(path,
                                           watcher,
                                           new GetZnodeChildrenCallback(),
                                           null);
                        }
                        catch (Exception ex) {
                            log.error("asyncGetZnodeChildren: Exception", ex);
                        }
                        break;
                    }
                    default:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc) + ", Failed !!!");
                        signal.countDown();
                        break;
                    }
                }
            }
        }
        
        zk.getChildren(path,
                       watcher,
                       new GetZnodeChildrenCallback(),
                       null);
        
        try {
            signal.await();
        }
        catch (InterruptedException ex) {
        }
        
        return children;
    }
    
    // ----- Get Data from Existing Znode -----
    
    final static String getDataZnode(ZooKeeper zk, String path) {
        String data = null;
        
        try {
            Stat stat = new Stat();
            
            byte[] ba = zk.getData(path,
                                   false,
                                   stat);
            if (ba != null) {
                data = new String(ba);
            }
        }
        catch (InterruptedException | KeeperException ex) {
            log.error("getDataZnode: Exception", ex);
        }
        
        return data;
    }
    
    final static String asyncGetDataZnode(ZooKeeper zk, String path) {
        CountDownLatch signal = new CountDownLatch(1);
        
        class GetDataZnodeContext {
            String data = null;
            
            void setData(String data) {
                this.data = data;
            }
        }
        
        final GetDataZnodeContext ctx = new GetDataZnodeContext();
        
        class GetZnodeDataCallback implements DataCallback {
            @Override
            public void processResult(int rc, String path, Object ctx, byte[] data, Stat stat)  {
                switch(KeeperException.Code.get(rc)) {
                    case OK:
                    {
                        ((GetDataZnodeContext)ctx).setData(new String(data));
                        log.info("ZK[" + path + "] -> Exists, Data = " + new String(data));
                        signal.countDown();
                        break;
                    }
                    case NONODE:
                    {
                        log.info("ZK[" + path + "] -> Does not Exists !!!");
                        signal.countDown();
                        break;
                    }
                    case CONNECTIONLOSS:
                    case SESSIONEXPIRED:
                    case SESSIONMOVED:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc));
                        try {
                            zk.getData(path,
                                       false,
                                       new GetZnodeDataCallback(),
                                       ctx);
                        }
                        catch (Exception ex) {
                            log.error("asyncGetDataZnode: Exception", ex);
                        }
                        break;
                    }
                    default:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc) + ", Failed !!!");
                        signal.countDown();
                        break;
                    }
                }
            }
        }
    
        zk.getData(path,
                   false,
                   new GetZnodeDataCallback(),
                   ctx);
        
        try {
            signal.await();
        }
        catch (InterruptedException ex) {
        }
        
        return ctx.data;
    }
    
    // ----- Set Data for an Existing Znode -----
    
    final static boolean setDataZnode(ZooKeeper zk, String path, byte[] data) {
        boolean success = false;
        
        try {
            Stat stat = zk.setData(path,
                                   data,
                                   -1);
            if (stat != null) {
                success = true;
            }
        }
        catch (InterruptedException | KeeperException ex) {
            log.error("setDataZnode: Exception", ex);
        }
        
        return success;
    }
    
    final static boolean asyncSetDataZnode(ZooKeeper zk, String path, byte[] data) {
        CountDownLatch signal = new CountDownLatch(1);
        
        class SetZnodeDataContext {
            boolean success = false;
            
            void setSuccess() {
                success = true;
            }
        }
        
        final SetZnodeDataContext ctx = new SetZnodeDataContext();
        
        class SetZnodeDataCallback implements StatCallback {
            @Override
            public void processResult(int rc, String path, Object ctx, Stat stat) {
                switch(KeeperException.Code.get(rc)) {
                    case OK:
                    {
                        ((SetZnodeDataContext)ctx).setSuccess();
                        log.info("ZK[" + path + "] -> Exists, Stat: {ctime = " + new Date(stat.getCtime())
                            + ", mtime = " + new Date(stat.getMtime())
                            + ", version = " + stat.getVersion() + "}");
                        signal.countDown();
                        break;
                    }
                    case NONODE:
                    {
                        log.info("ZK[" + path + "] -> Does not Exists !!!");
                        signal.countDown();
                        break;
                    }
                    case CONNECTIONLOSS:
                    case SESSIONEXPIRED:
                    case SESSIONMOVED:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc));
                        try {
                            zk.setData(path,
                                       data,
                                       -1,
                                       new SetZnodeDataCallback(),
                                       ctx);
                        }
                        catch (Exception ex) {
                            log.error("asyncSetDataZnode: Exception", ex);
                        }
                        break;
                    }
                    default:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc) + ", Failed !!!");
                        signal.countDown();
                        break;
                    }
                }
            }
        }
    
        zk.setData(path,
                   data,
                   -1,
                   new SetZnodeDataCallback(),
                   ctx);
        
        try {
            signal.await();
        }
        catch (InterruptedException ex) {
        }
        
        return ctx.success;
    }
    
    // ----- Delete an Existing Znode -----
    
    final static boolean deleteZnode(ZooKeeper zk, String path) {
        boolean success = false;
        
        try {
            zk.delete(path,
                      -1);
            
            success = true;
        }
        catch (InterruptedException | KeeperException ex) {
            log.error("deleteZnode: Exception", ex);
        }
        
        return success;
    }
    
    final static boolean asyncDeleteZnode(ZooKeeper zk, String path) {
        CountDownLatch signal = new CountDownLatch(1);
        
        class DeleteZnodeContext {
            boolean success = false;
            
            void setSuccess() {
                success = true;
            }
        }
        
        final DeleteZnodeContext ctx = new DeleteZnodeContext();
        
        class DeleteZnodeVoidCallback implements VoidCallback {
            @Override
            public void processResult(int rc, String path, Object ctx) {
                switch(KeeperException.Code.get(rc)) {
                    case OK:
                    {
                        ((DeleteZnodeContext)ctx).setSuccess();
                        log.info("ZK[" + path + "] -> Deleted !!!");
                        signal.countDown();
                        break;
                    }
                    case NONODE:
                    {
                        log.info("ZK[" + path + "] -> Does not Exists !!!");
                        signal.countDown();
                        break;
                    }
                    case CONNECTIONLOSS:
                    case SESSIONEXPIRED:
                    case SESSIONMOVED:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc));
                        try {
                            zk.delete(path,
                                      -1,
                                      new DeleteZnodeVoidCallback(),
                                      ctx);
                        }
                        catch (Exception ex) {
                            log.error("asyncDeleteZnode: Exception", ex);
                        }
                        break;
                    }
                    default:
                    {
                        log.info("ZK[" + path + "] -> Code = " + KeeperException.Code.get(rc) + ", Failed !!!");
                        signal.countDown();
                        break;
                    }
                }
            }
        }
        
        zk.delete(path,
                  -1,
                  new DeleteZnodeVoidCallback(),
                  ctx);
        
        try {
            signal.await();
        }
        catch (InterruptedException ex) {
        }
        
        return ctx.success;
    }
    
    // ----- Close Connection to ZooKeeper -----
    
    final static void close(ZooKeeper zk)
        throws Exception {
        zk.close();
    }
}
